from __future__ import annotations

from jinja2 import ChoiceLoader, Environment, FileSystemLoader, PackageLoader


def render_template(name: str, **kwargs) -> str:
    """
    Render a Jinja template
    """
    loader = ChoiceLoader(
        [
            FileSystemLoader("."),
            PackageLoader("emotional", "templates"),
        ]
    )
    env = Environment(loader=loader, trim_blocks=True)
    jinja_template = env.get_template(name)
    return jinja_template.render(**kwargs)



from __future__ import annotations

import pytest
from commitizen import bump
from commitizen.git import GitCommit

from emotional.plugin import Emotional

NONE_INCREMENT = [
    "docs(README): motivation",
    "ci: added travis",
    "performance. Remove or disable the reimplemented linters",
    "refactor that how this line starts",
]

PATCH_INCREMENTS = [
    "fix(setup.py): future is now required for every python version",
    "docs(README): motivation",
]

MINOR_INCREMENTS = [
    "feat(cli): added version",
    "docs(README): motivation",
    "fix(setup.py): future is now required for every python version",
    "perf: app is much faster",
    "refactor: app is much faster",
]

MAJOR_INCREMENTS_BREAKING_CHANGE = [
    "feat(cli): added version",
    "docs(README): motivation",
    "BREAKING CHANGE: `extends` key in config file is now used for extending other config files",  # noqa
    "fix(setup.py): future is now required for every python version",
]

MAJOR_INCREMENTS_BREAKING_CHANGE_ALT = [
    "feat(cli): added version",
    "docs(README): motivation",
    "BREAKING-CHANGE: `extends` key in config file is now used for extending other config files",  # noqa
    "fix(setup.py): future is now required for every python version",
]

MAJOR_INCREMENTS_EXCLAMATION = [
    "feat(cli)!: added version",
    "docs(README): motivation",
    "fix(setup.py): future is now required for every python version",
]

MAJOR_INCREMENTS_EXCLAMATION_SAMPLE_2 = ["feat(pipeline)!: some text with breaking change"]

MAJOR_INCREMENTS_EXCLAMATION_OTHER_TYPE = [
    "chore!: drop support for Python 3.9",
    "docs(README): motivation",
    "fix(setup.py): future is now required for every python version",
]


@pytest.mark.parametrize(
    "messages, expected_type",
    (
        (PATCH_INCREMENTS, "PATCH"),
        (MINOR_INCREMENTS, "MINOR"),
        (MAJOR_INCREMENTS_BREAKING_CHANGE, "MAJOR"),
        (MAJOR_INCREMENTS_BREAKING_CHANGE_ALT, "MAJOR"),
        (MAJOR_INCREMENTS_EXCLAMATION_OTHER_TYPE, "MAJOR"),
        (MAJOR_INCREMENTS_EXCLAMATION, "MAJOR"),
        (MAJOR_INCREMENTS_EXCLAMATION_SAMPLE_2, "MAJOR"),
        (NONE_INCREMENT, None),
    ),
)
def test_find_increment(messages, expected_type, config):
    cz = Emotional(config)
    commits = [GitCommit(rev="test", title=message) for message in messages]
    increment_type = bump.find_increment(
        commits,
        regex=cz.bump_pattern,
        increments_map=cz.bump_map,
    )
    assert increment_type == expected_type


@pytest.mark.parametrize(
    "messages, expected_type",
    (
        (PATCH_INCREMENTS, "PATCH"),
        (MINOR_INCREMENTS, "MINOR"),
        (MAJOR_INCREMENTS_BREAKING_CHANGE, "MINOR"),
        (MAJOR_INCREMENTS_BREAKING_CHANGE_ALT, "MINOR"),
        (MAJOR_INCREMENTS_EXCLAMATION_OTHER_TYPE, "MINOR"),
        (MAJOR_INCREMENTS_EXCLAMATION, "MINOR"),
        (MAJOR_INCREMENTS_EXCLAMATION_SAMPLE_2, "MINOR"),
        (NONE_INCREMENT, None),
    ),
)
def test_find_increment_major_version_zero(messages, expected_type, config):
    cz = Emotional(config)
    commits = [GitCommit(rev="test", title=message) for message in messages]
    increment_type = bump.find_increment(
        commits,
        regex=cz.bump_pattern,
        increments_map=cz.bump_map_major_version_zero,
    )
    assert increment_type == expected_type



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from emotional.config import EmotionalConfig

if TYPE_CHECKING:
    from _pytest.mark import ParameterSet


def param(id, url, repo, **settings) -> ParameterSet:
    return pytest.param(url, repo, marks=pytest.mark.settings(**settings), id=id)


@pytest.mark.parametrize(
    "url,repository",
    [
        param("repo", "https://github.com", "org/repo", github="org/repo"),
        param("url", "https://github.com", "org/repo", github="https://github.com/org/repo"),
        param("hosted", "https://private.com", "org/repo", github="https://private.com/org/repo"),
    ],
)
def test_config_github_repository(emotional_config: EmotionalConfig, url: str, repository: str):
    assert emotional_config.github == repository
    assert emotional_config.github_url == url


@pytest.mark.parametrize(
    "url,repository",
    [
        param("repo", "https://gitlab.com", "org/repo", gitlab="org/repo"),
        param("url", "https://gitlab.com", "org/repo", gitlab="https://gitlab.com/org/repo"),
        param("hosted", "https://private.com", "org/repo", gitlab="https://private.com/org/repo"),
    ],
)
def test_config_gitlab_repository(emotional_config: EmotionalConfig, url: str, repository: str):
    assert emotional_config.gitlab == repository
    assert emotional_config.gitlab_url == url



from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files  # type: ignore[no-redef,import-not-found]


def read_version() -> str:
    try:
        return version("emotional")
    except (PackageNotFoundError, ImportError):
        version_file = files(__package__) / "VERSION"
        return version_file.read_text() if version_file.is_file() else "0.0.0.dev"


__version__ = read_version()



"""
See:
    - https://docs.github.com/en/issues/tracking-your-work-with-issues/linking-a-pull-request-to-an-issue
    - https://docs.github.com/en/get-started/writing-on-github/working-with-advanced-formatting/autolinked-references-and-urls#issues-and-pull-requests
"""  # noqa: E501
from __future__ import annotations

import re

from commitizen.git import GitCommit

from .config import EmotionalConfig

RE_ISSUE = re.compile(r"(?P<repository>(?P<owner>\w+)/(?P<project>\w+))?#(?P<issue>\d+)")

KEYWORDS = (
    "close",
    "closes",
    "closed",
    "fix",
    "fixes",
    "fixed",
    "resolve",
    "resolves",
    "resolved",
)


def changelog_message_hook(
    config: EmotionalConfig, parsed_message: dict, commit: GitCommit
) -> dict:
    if config.github is None:
        return parsed_message

    def urlize_issue(match: re.Match[str]) -> str:
        label = match.group(0)
        repository = match.group("repository") or config.github
        issue = match.group("issue")
        return f"[{label}]({config.github_url}/{repository}/issues/{issue})"

    for field in "message", "scope":
        value = parsed_message[field]
        if value:
            parsed_message[field] = RE_ISSUE.sub(urlize_issue, value)

    for field in "body", "footers":
        body = parsed_message.get(field)
        if body:
            for match in RE_ISSUE.finditer(body):
                parsed_message["message"] += f" {urlize_issue(match)}"

    return parsed_message



"""
See:
    - https://support.atlassian.com/jira-software-cloud/docs/process-issues-with-smart-commits/
    - https://confluence.atlassian.com/fisheye/using-smart-commits-960155400.html
"""
from __future__ import annotations

import re

from commitizen.git import GitCommit

from .config import EmotionalConfig


def changelog_message_hook(
    config: EmotionalConfig, parsed_message: dict, commit: GitCommit
) -> dict:
    if not config.jira_url or not config.jira_prefixes:
        return parsed_message

    prefixes = "|".join(config.jira_prefixes)
    re_issue = re.compile(rf"({prefixes})\d+")

    def urlize_issue(match: re.Match[str]) -> str:
        issue = match.group(0)
        return f"[{issue}]({config.jira_url}/browse/{issue})"

    for field in "message", "scope":
        value = parsed_message[field]
        if value:
            parsed_message[field] = re_issue.sub(urlize_issue, value)

    for field in "body", "footers":
        body = parsed_message.get(field)
        if body:
            for match in re_issue.finditer(body):
                parsed_message["message"] += f" {urlize_issue(match)}"

    return parsed_message



from __future__ import annotations

RELEASE_EMOJI = "🚀"

TYPES: list[dict] = [
    dict(
        type="BREAKING CHANGE",
        description="Changes that are not backward-compatibles",
        heading="Breaking changes",
        emoji="🚨",
        bump="MAJOR",
        regex=r"BREAKING[\-\ ]CHANGE",
        question=False,  # Breaking changes have a dedicated question
    ),
    dict(
        type="feat",
        description="A new feature",
        heading="New features",
        emoji="💫",
        bump="MINOR",
        key="n",
    ),
    dict(
        type="fix",
        description="A bug fix",
        heading="Bug fixes",
        emoji="🐛",
        bump="PATCH",
    ),
    dict(
        type="perf",
        description="A changeset improving performance",
        heading="Performance",
        emoji="📈",
        aliases=["performance"],
        bump="PATCH",
    ),
    dict(
        type="docs",
        description="Documentation only change",
        heading="Documentation",
        emoji="📖",
        aliases=["doc"],
    ),
    dict(
        type="build",
        description="Changes that affect the build system or external dependencies (ex: pip, docker, npm)",
        heading="Build",
        emoji="📦",
        aliases=["deps"],
    ),
    dict(
        type="style",
        description="Changes that do not affect the meaning of the code (white-space, formatting, …)",
        heading="Style",
        emoji="🎨",
        changelog=False,
    ),
    dict(
        type="test",
        description="Adding missing or correcting existing tests",
        heading="Testing",
        emoji="🚦",
        aliases=["tests"],
        changelog=False,
    ),
    dict(
        type="ci",
        description="Changes to CI configuration files and scripts",
        heading="Continous Integration",
        emoji="🛸",
        changelog=False,
    ),
    dict(
        type="refactor",
        description="A changeset neither fixing a bug nor adding a feature",
        heading="Refactorings",
        emoji="🔧",
        changelog=False,
        bump="PATCH",
    ),
    dict(
        type="i18n",
        description="A changeset related to languages and translations",
        heading="Internationalization",
        emoji="🌍",
        aliases=["locales", "l10n"],
        bump="PATCH",
    ),
    dict(
        type="chore",
        description="Changes not fitting in other categories",
        heading="Chores",
        emoji="🧹",
        key="o",
    ),
    dict(
        type="revert",
        description="Revert one or more commits",
        heading="Reverted",
        emoji="🔙",
        changelog=False,
        key="e",
    ),
    dict(
        type="wip",
        description="Work in progress",
        heading="Work in progress",
        emoji="🚧",
        changelog=False,
    ),
    dict(
        type="bump",
        description="A bump commit",
        heading="",
        emoji=RELEASE_EMOJI,
        aliases=["release"],
        changelog=False,
        question=False,
    ),
]



import pytest
from commitizen.commands.check import Check
from commitizen.cz.exceptions import AnswerRequiredError

from emotional.plugin import Emotional, parse_scope, parse_subject

valid_scopes = ["", "simple", "dash-separated", "camelCase" "UPPERCASE"]

scopes_transformations = [["with spaces", "with-spaces"], [None, ""]]

valid_subjects = ["this is a normal text", "aword"]

subjects_transformations = [["with dot.", "with dot"]]

invalid_subjects = ["", "   ", ".", "   .", "", None]


def test_parse_scope_valid_values():
    for valid_scope in valid_scopes:
        assert valid_scope == parse_scope(valid_scope)


def test_scopes_transformations():
    for scopes_transformation in scopes_transformations:
        invalid_scope, transformed_scope = scopes_transformation
        assert transformed_scope == parse_scope(invalid_scope)


def test_parse_subject_valid_values():
    for valid_subject in valid_subjects:
        assert valid_subject == parse_subject(valid_subject)


def test_parse_subject_invalid_values():
    for valid_subject in invalid_subjects:
        with pytest.raises(AnswerRequiredError):
            parse_subject(valid_subject)


def test_subject_transformations():
    for subject_transformation in subjects_transformations:
        invalid_subject, transformed_subject = subject_transformation
        assert transformed_subject == parse_subject(invalid_subject)


def test_questions(config):
    emotional = Emotional(config)
    questions = emotional.questions()
    assert isinstance(questions, list)
    assert isinstance(questions[0], dict)


def test_choices_all_have_keyboard_shortcuts(config):
    emotional = Emotional(config)
    questions = emotional.questions()

    list_questions = (q for q in questions if q["type"] == "list")
    for select in list_questions:
        assert all("key" in choice for choice in select["choices"])


def test_choices_dont_have_duplicate_keyboard_shortcuts(config):
    emotional = Emotional(config)
    questions = emotional.questions()

    list_questions = (q for q in questions if q["type"] == "list")
    for select in list_questions:
        shortcuts = [choice.get("key") for choice in select["choices"]]
        assert len(set(shortcuts)) == len(shortcuts)


def test_small_answer(config):
    emotional = Emotional(config)
    answers = {
        "prefix": "fix",
        "scope": "users",
        "subject": "email pattern corrected",
        "is_breaking_change": False,
        "body": "",
        "footer": "",
    }
    message = emotional.message(answers)
    assert message == "fix(users): email pattern corrected"


def test_no_scope(config):
    emotional = Emotional(config)
    answers = {
        "prefix": "fix",
        "scope": "",
        "subject": "email pattern corrected",
        "is_breaking_change": False,
        "body": "",
        "footer": "",
    }
    message = emotional.message(answers)
    assert message == "fix: email pattern corrected"


def test_long_answer(config):
    emotional = Emotional(config)
    answers = {
        "prefix": "fix",
        "scope": "users",
        "subject": "email pattern corrected",
        "is_breaking_change": False,
        "body": "complete content",
        "footer": "closes #24",
    }
    message = emotional.message(answers)
    assert message == (
        "fix(users): email pattern corrected\n" "\n" "complete content\n" "\n" "closes #24"  # noqa
    )


def test_breaking_change_in_footer(config):
    emotional = Emotional(config)
    answers = {
        "prefix": "fix",
        "scope": "users",
        "subject": "email pattern corrected",
        "body": "complete content",
        "is_breaking_change": True,
        "breaking_change": "breaking change content",
        "footer": "Fixes #42",
    }
    message = emotional.message(answers)
    assert message == (
        "fix(users): email pattern corrected\n"
        "\n"
        "complete content\n"
        "\n"
        "BREAKING CHANGE: breaking change content\n"
        "Fixes #42"
    )


def test_exclamation_mark_breaking_change(config):
    emotional = Emotional(config)
    answers = {
        "prefix": "fix",
        "scope": "users",
        "subject": "email pattern corrected",
        "body": "complete content",
        "is_breaking_change": True,
        "breaking_change": "",
        "footer": "Fixes #42",
    }
    message = emotional.message(answers)
    assert message == (
        "fix(users)!: email pattern corrected\n" "\n" "complete content\n" "\n" "Fixes #42"
    )


def test_exclamation_mark_breaking_change_without_scope(config):
    emotional = Emotional(config)
    answers = {
        "prefix": "fix",
        "scope": "",
        "subject": "email pattern corrected",
        "body": "complete content",
        "is_breaking_change": True,
        "breaking_change": "",
        "footer": "Fixes #42",
    }
    message = emotional.message(answers)
    assert message == ("fix!: email pattern corrected\n" "\n" "complete content\n" "\n" "Fixes #42")


@pytest.mark.parametrize(
    "message",
    [
        "bump: version 0.0.0 → 0.1.0",
        "bump: version 0.1.0 → 0.1.1",
        "bump: version 1.0.0 → 1.1.0",
    ],
)
def test_validate_bump_commit(config, message: str):
    emotional = Emotional(config)
    check = Check(config, {"message": message})
    check.validate_commit_message(message, emotional.schema_pattern())
